
from sqlite3 import connect, Row

from ch2.command.args import bootstrap_file, m, V
from ch2.config.personal import acooke
from ch2.lib.date import to_time
from ch2.squeal.database import add
from ch2.squeal.tables.source import Source
from ch2.squeal.tables.statistic import StatisticJournal, StatisticJournalText, StatisticJournalInteger, \
    StatisticJournalFloat, StatisticName
from ch2.squeal.tables.topic import TopicJournal, Topic

old = connect('/home/andrew/.ch2/database.sqld')
old.row_factory = Row


class File:
    name = '/home/andrew/.ch2/database.sqle'


args, log, db = bootstrap_file(File(), m(V), '5', configurator=acooke)
s = db.session()


def assert_empty(cls):
    assert s.query(cls).count() == 0, cls


assert_empty(Source)
assert_empty(StatisticJournal)
assert_empty(TopicJournal)


diary = s.query(Topic).filter(Topic.name == 'DailyDiary').one()
fields = dict((field.statistic_name.name, field.statistic_name) for field in diary.fields)
notes = fields['Notes']
mood = fields['Mood']
hr = fields['Rest HR']
weight = fields['Weight']
sleep = fields['Sleep']
weather = fields['Weather']
meds = fields['Medication']

for row in old.execute('''select date, notes, rest_heart_rate, sleep, mood, weather, medication, weight from diary''', []):
    if row['notes'] or row['mood'] or row['rest_heart_rate'] or row['weight'] or row['sleep'] or row['weather']:
        tj = add(s, TopicJournal(time=to_time(row['date']), topic=diary))
        if row['notes']:
            add(s, StatisticJournalText(statistic_name=notes, source=tj, value=row['notes']))
        if row['mood']:
            add(s, StatisticJournalInteger(statistic_name=mood, source=tj, value=row['mood']))
        if row['rest_heart_rate']:
            add(s, StatisticJournalInteger(statistic_name=hr, source=tj, value=row['rest_heart_rate']))
        if row['weight']:
            add(s, StatisticJournalFloat(statistic_name=weight, source=tj, value=row['weight']))
        if row['sleep']:
            add(s, StatisticJournalFloat(statistic_name=sleep, source=tj, value=row['sleep']))
        if row['weather']:
            add(s, StatisticJournalText(statistic_name=weather, source=tj, value=row['weather']))
        if row['medication']:
            add(s, StatisticJournalText(statistic_name=meds, source=tj, value=row['medication']))


def injury_notes(old_name, new_name):
    injury_id = next(old.execute('''select id from injury where name like ?''', [old_name]))[0]
    topic = s.query(Topic).filter(Topic.name == new_name).one()
    notes = s.query(StatisticName).filter(StatisticName.name == 'Notes', StatisticName.constraint == topic.id).one()
    for row in old.execute('''select date, notes from injury_diary where injury_id = ?''', [injury_id]):
        if row['notes']:
            # print(row['notes'], len(row['notes']))
            tj = add(s, TopicJournal(time=to_time(row['date']), topic=topic))
            add(s, StatisticJournalText(statistic_name=notes, source=tj, value=row['notes']))


injury_notes('MS (General Notes)', 'Multiple Sclerosis')
injury_notes('Tendon pain (femur, lhs)', 'Broken Femur LHS')

s.flush()
s.commit()
s.close()

print('end')




from collections import defaultdict

from pandas import DataFrame
from sqlalchemy import inspect, select
from sqlalchemy.sql.functions import count, coalesce

from ..squeal.database import connect
from ..squeal.tables.activity import ActivityGroup, ActivityJournal
from ..squeal.tables.monitor import MonitorJournal
from ..squeal.tables.segment import Segment, SegmentJournal
from ..squeal.tables.source import Interval, Source
from ..squeal.tables.statistic import StatisticName, StatisticJournal, StatisticMeasure, StatisticJournalInteger, \
    StatisticJournalFloat, StatisticJournalText


def extract(data, instance, *attributes):
    for attribute in attributes:
        if hasattr(instance, attribute):
            data[attribute].append(getattr(instance, attribute))
        else:
            data[attribute].append(None)


class Data:
    '''
    Provide access to the database via DataFrames.
    Intended for use in Jupyter notebooks.
    All data retrieved in a single (read-only) session.
    '''

    def __init__(self, log, db):
        self._log = log
        self._s = db.session()

    def activity_groups(self):
        data, ids = defaultdict(list), []
        for activity_group in self._s.query(ActivityGroup).order_by(ActivityGroup.name):
            ids.append(activity_group.id)
            extract(data, activity_group, 'name', 'description', 'id')
            data['count'].append(self._s.query(count(ActivityJournal.id)).
                                 filter(ActivityJournal.activity_group == activity_group).scalar())
        return DataFrame(data, index=ids)

    def activity_journals(self, group, start=None, finish=None):
        data, times = defaultdict(list), []
        q = self._s.query(ActivityJournal).join(ActivityGroup).filter(ActivityGroup.name == group)
        if start:
            q = q.filter(ActivityJournal.finish >= start)
        if finish:
            q = q.filter(ActivityJournal.start <= finish)
        self._log.debug(q)
        for journal in q.order_by(ActivityJournal.start).all():
            times.append(journal.start)
            extract(data, journal, 'id', 'name', 'fit_file', 'finish')
        data['source_id'] = data['id']; del data['id']
        return DataFrame(data, index=times)

    def statistic_names(self, *names):
        statistic_names, statistic_ids = self._collect_statistics(names)
        data, ids = defaultdict(list), []
        for statistic in self._s.query(StatisticName). \
                filter(StatisticName.id.in_(statistic_ids)).order_by(StatisticName.name):
            ids.append(statistic.id)
            extract(data, statistic, 'name', 'description', 'units', 'summary', 'owner', 'constraint', 'id')
            data['count'].append(self._s.query(count(StatisticJournal.id)).
                                 filter(StatisticJournal.statistic_name == statistic).scalar())
        return DataFrame(data, index=ids)

    def _collect_statistics(self, names):
        if not names:
            names = ['%']
        statistic_ids, statistic_names = set(), set()
        for statistic_name in names:
            for statistic_name in self._s.query(StatisticName). \
                    filter(StatisticName.name.like(statistic_name)).all():
                statistic_ids.add(statistic_name.id)
                statistic_names.add(statistic_name.name)
        return statistic_names, statistic_ids

    def _build_statistic_journal_query(self, statistic_ids, start, finish, owner, constraint, source_ids, schedule):

        # use more efficient expression interface and exploit the fact that
        # alternative journal types will be null in an outer join.

        sj = inspect(StatisticJournal).local_table
        sn = inspect(StatisticName).local_table
        sji = inspect(StatisticJournalInteger).local_table
        sjf = inspect(StatisticJournalFloat).local_table
        sjt = inspect(StatisticJournalText).local_table
        inv = inspect(Interval).local_table

        q = select([sn.c.name, sj.c.time, coalesce(sjf.c.value, sji.c.value, sjt.c.value)]). \
            select_from(sj.join(sn).outerjoin(sjf).outerjoin(sji).outerjoin(sjt)). \
            where(sn.c.id.in_(statistic_ids))
        if start:
            q = q.where(sj.c.time >= start)
        if finish:
            q = q.where(sj.c.time <= finish)
        if owner:
            q = q.where(sn.c.owner == owner)
        if constraint:
            q = q.where(sn.c.constraint == constraint)
        if source_ids is not None:  # avoid testing DataFrame sequences as bools
            q = q.where(sj.c.source_id.in_(source_ids))
        if schedule:
            q = q.join(inv).where(inv.c.schedule == schedule)
        q = q.order_by(sj.c.time)
        return q

    def statistic_journals(self, *statistics,
                           start=None, finish=None, owner=None, constraint=None, source_ids=None, schedule=None):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        q = self._build_statistic_journal_query(statistic_ids, start, finish, owner, constraint, source_ids, schedule)
        self._log.debug(q)
        data, times, err_cnt = defaultdict(list), [], defaultdict(lambda: 0)

        def pad():
            n = len(times)
            for name in statistic_names:
                if len(data[name]) != n:
                    err_cnt[name] += 1
                    if err_cnt[name] <= 1:
                        self._log.warn('Missing %s at %s (only warning for this name)' % (name, times[-1]))
                    data[name].append(None)

        for name, time, value in self._s.connection().execute(q):
            if times and times[-1] != time:
                pad()
            if not times or times[-1] != time:
                times.append(time)
            if len(data[name]) >= len(times):
                raise Exception('Duplicate data for %s at %s ' % (name, time) +
                                '(you may need to specify more constraints to make the query unique)')
            data[name].append(value)
        pad()
        self._log.debug('Loaded %d distinct times' % len(times))
        return DataFrame(data, index=times)

    def _build_statistic_quartiles_query(self, statistic_ids, q, start, finish, owner, constraint, source_id):
        q = q.filter(StatisticName.id.in_(statistic_ids))
        if start:
            q = q.filter(StatisticJournal.time >= start)
        if finish:
            q = q.filter(StatisticJournal.time <= finish)
        if owner:
            q = q.filter(StatisticName.owner == owner)
        if constraint:
            q = q.filter(StatisticName.constraint == constraint)
        if source_id:
            q = q.filter(StatisticJournal.source_id == int(source_id))
        return q

    def statistic_quartiles(self, *statistics, schedule='m',
                            start=None, finish=None, owner=None, constraint=None, source_ids=None):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        q = self._s.query(StatisticMeasure). \
            join(StatisticJournal, StatisticMeasure.statistic_journal_id == StatisticJournal.id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            join(Source, StatisticJournal.source_id == Source.id). \
            filter(StatisticName.id.in_(statistic_ids)). \
            filter(StatisticMeasure.quartile != None)
        if start:
            q = q.filter(StatisticJournal.time >= start)
        if finish:
            q = q.filter(StatisticJournal.time <= finish)
        if owner:
            q = q.filter(StatisticName.owner == owner)
        if constraint:
            q = q.filter(StatisticName.constraint == constraint)
        if source_ids is not None:
            q = q.filter(StatisticJournal.source_id.in_(source_ids))
        if schedule:
            q = q.join((Interval, StatisticMeasure.source_id == Interval.id)). \
                filter(Interval.schedule == schedule)
        self._log.debug(q)
        raw_data = defaultdict(lambda: defaultdict(lambda: [0] * 5))
        for measure in q.all():
            raw_data[measure.source.start][measure.statistic_journal.statistic_name.name][measure.quartile] = \
                measure.statistic_journal.value
        data, times = defaultdict(list), []
        for time in sorted(raw_data.keys()):
            times.append(time)
            sub_data = raw_data[time]
            for statistic in statistic_names:
                if statistic in sub_data:
                    data[statistic].append(sub_data[statistic])
                else:
                    data[statistic].append(None)
        return DataFrame(data, index=times)

    def monitor_journals(self):
        data, times = defaultdict(list), []
        for journal in self._s.query(MonitorJournal).order_by(MonitorJournal.start).all():
            times.append(journal.start)
            extract(data, journal, 'id', 'fit_file', 'finish')
        data['source_id'] = data['id']; del data['id']
        return DataFrame(data, index=times)

    def segments(self):
        data, ids = defaultdict(list), []
        for segment in self._s.query(Segment).order_by(Segment.name):
            ids.append(segment.id)
            extract(data, segment, 'name', 'description', 'id')
            data['count'].append(self._s.query(count(SegmentJournal.id)).
                                 filter(SegmentJournal.segment == segment).scalar())
        return DataFrame(data, index=ids)

    def segment_journals(self):
        data, times = defaultdict(list), []
        for journal in self._s.query(SegmentJournal).order_by(SegmentJournal.start).all():
            times.append(journal.start)
            extract(data, journal, 'id', 'start', 'finish', 'activity_journal_id', 'segment_id')
        data['source_id'] = data['id']; del data['id']
        return DataFrame(data, index=times)


def data(*args):
    '''
    Start here to access data.  Create an instance in Jupyter:

        from ch2.data import data
        d = data('-v', '4')
        d.statistics()
        ...
    '''
    ns, log, db = connect(args)
    return Data(log, db)




from time import sleep

from .args import DIR, USER, PASS, DATE
from ..fit.download.connect import GarminConnect
from ..stoats.read.monitor import missing_dates


def garmin(args, log, db):
    '''
## garmin

    > ch2 garmin --user USER --pass PASSWORD DIR

Download recent monitor data to the given directory.

    > ch2 garmin --user USER --pass PASSWORD --date DATE DIR

Download monitor data for the given date.

Note that this cannot be used to download more than 10 days of data.
For bulk downloads use
https://www.garmin.com/en-US/account/datamanagement/
    '''
    dir, user, password, date = args.dir(DIR, rooted=False), args[USER], args[PASS], args[DATE]
    if date:
        dates = [date]
    else:
        # do this first to avoid login if not needed
        with db.session_context() as s:
            dates = list(missing_dates(log, s))
    if dates:
        connect = GarminConnect(log, log_response=False)
        connect.login(user, password)
        for repeat, date in enumerate(dates):
            if repeat:
                sleep(1)
            log.info('Downloading data for %s' % date)
            connect.get_monitoring_to_fit_file(date, dir)




import datetime as dt
from abc import ABC, abstractmethod
from calendar import monthrange
from re import sub, compile, match

from .date import to_date, format_date, add_date, local_date_to_time, time_to_local_date

# my calculations are done relative to the unix epoch.  the "gregorian ordinal"
# is relative to year 1, but i have no idea how the details of that work.  i
# guess i could do everything in gregorian ordinals simply projecting back the
# current weeks / months and it would be equivalent, but i'd need to tweak the
# week offset by hand (here it's because 1970-01-01 is a thursday).

ZERO = dt.date(1970, 1, 1)
WEEK_OFFSET = 3
EPOCH_OFFSET = ZERO.toordinal()
DOW = ('mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun')


class Schedule:
    """
    Parse a spec and reduce it to a normalized form (available through __str__()).

    Offset can shift whole "units" only.  So 3w can start on *any* Monday,
    depending on the offset data, but the week always starts on a Monday.

    Note that dates are dates, not datetimes (times).
    """

    DOW_INDEX = dict((day, i) for i, day in enumerate(DOW))

    def __init__(self, spec):
        self.__start = None  # dt.date visible via self.start (inclusive)
        self.__finish = None  # dt.date visible via self.finish (exclusive)
        self.offset = None  # int (units of frame relative to start of unix epoch)
        self.repeat = None  # int (units of frame)
        self.frame_type = None  # character (as spec, so all lower case)
        self.locations = None  # list of day offsets or (week, dow) tuples (if empty, all dates)
        try:
            spec = '' if spec is None else spec
            spec = sub(r'\s+', '', spec)
            spec = spec.lower()
            m = match(r'^([\d\-/]*[dwmy])?(\[[^\]]*\])?([\d\-]*)$', spec)
            frame, locations, range = m.group(1), m.group(2), m.group(3)
            self.__parse_frame(frame)
            self.__parse_locations(locations)
            self.__parse_range(range)
        except:
            raise Exception('Cannot parse "%s" (%s)' % (spec, type(spec)))
        if self.locations and self.frame_type == 'y':
            raise Exception('Locations not supported in yearly schedules')
        self.__frame = self.frame_class()(self)

    def __date_to_ordinal(self, text):
        return DateOrdinals(text).ordinals[self.frame_type]

    @staticmethod
    def __parse_ordinal(text):
        if text.isdecimal():
            return int(text), 'd'
        elif len(text) > 1:
            return int(text[:-1]), text[-1]
        else:
            return 1, text

    def __parse_frame(self, frame):
        if frame is None:
            frame = 'd'
        if '/' in frame:
            offset, rft = frame.split('/')
        else:
            offset, rft = '0', frame
        self.repeat, self.frame_type = self.__parse_ordinal(rft)
        if '-' in offset:
            self.offset = self.__date_to_ordinal(offset)
        else:
            self.offset = int(offset)
        self.offset %= self.repeat

    def __parse_location(self, location):
        if location[-3:] in DOW:
            if len(location) > 3:
                if self.frame_type == 'd':
                    raise Exception('Numbered locations in daily frames not supported')
                return int(location[:-3]), self.DOW_INDEX[location[-3:]]
            else:
                return 0, self.DOW_INDEX[location]  # 0 means all weeks (in a month)
        else:
            return int(location)

    def __parse_locations(self, locations):
        if locations:
            locations = locations[1:-1]  # drop []
        if locations:
            if self.frame_type == 'y':
                raise Exception('Locations in yearly frames not supported')
            self.locations = sorted(map(self.__parse_location, locations.split(',')), key=self.__key)
        else:
            self.locations = []  # all

    def __key(self, location):
        if isinstance(location, int):
            return (-1, location)
        else:
            return location

    def __parse_range(self, range):
        if range:
            m = compile(r'^(\d+-\d+-\d+)?-(\d+-\d+-\d+)?$').match(range)
            if m:
                self.__start = to_date(m.group(1)) if m.group(1) else None
                self.__finish = to_date(m.group(2)) if m.group(2) else None
            else:
                # if a single value is given then it is the start with an implicit finish that
                # is the day after (so a single day, since finish is exclusive)
                self.__start = to_date(range)
                self.__finish = self.__start + dt.timedelta(days=1)
        else:
            self.__start, self.__finish = None, None

    def __str__(self):
        return '%s%s%s' % (self.__str_offset(), self.__str_locations(), self.__str_ranges())

    def __str_offset(self):
        repeat = '%d' % self.repeat if self.repeat > 1 else ''
        if self.offset:
            return '%d/%s%s' % (self.offset, repeat, self.frame_type)
        elif repeat or self.frame_type != 'd' or not (self.start or self.finish):
            return '%s%s' % (repeat, self.frame_type)
        else:
            return ''

    def __str_locations(self):
        if self.locations:
            return '[%s]' % ','.join(map(self.__str_location, self.locations))
        else:
            return ''

    def __str_location(self, location):
        try:
            if location[0]:
                return '%d%s' % (location[0], DOW[location[1]])
            else:
                return DOW[location[1]]
        except:
            return str(location)

    def __str_ranges(self):
        if self.__start is None and self.__finish is None:
            return ''
        elif self.__start and self.__start + dt.timedelta(days=1) == self.__finish:
            return self.__str_range(self.__start)
        elif self.__start and not self.__finish:
            return '%s-' % self.__str_range(self.__start)
        elif self.__finish and not self.__start:
            return '-%s' % self.__str_range(self.__finish)
        else:
            return '%s-%s' % (self.__str_range(self.__start), self.__str_range(self.finish))

    def __str_range(self, range):
        if range is None:
            return ''
        else:
            return format_date(range)

    def frame_class(self):
        return {'d': Day, 'w': Week, 'm': Month, 'y': Year}[self.frame_type]

    # allow range to be set separately (allows separate column in database, so
    # we can select for valid reminders).

    def __set_start(self, date):
        self.__start = to_date(date, none=True)

    def __set_finish(self, date):
        self.__finish = to_date(date, none=True)

    start = property(lambda self: self.__start, __set_start)
    finish = property(lambda self: self.__finish, __set_finish)

    def in_range(self, date):
        date = to_date(date)
        return (self.start is None or date >= self.start) and \
               (self.finish is None or date < self.finish)

    def describe(self, compact=False):
        if compact:
            text = self.frame_type
        else:
            text = '%s' % self.frame_class().__name__
        if self.repeat > 1:
            text = '%d%ss' % (self.repeat, text)
        text = '%s%s' % (text, self.__str_locations())
        return text

    def __in_range_or_none(self, date):
        if self.in_range(date):
            return date
        else:
            return None

    def start_of_frame(self, date):
        return self.__in_range_or_none(self.__frame.start_of_frame(date))

    def next_frame(self, date):
        return self.__in_range_or_none(self.__frame.start_of_frame(date) +
                                       dt.timedelta(days=self.frame_length_in_days(date)))

    def frame_length_in_days(self, date):
        return self.__frame.length_in_days(date)

    def locations_from(self, start):
        '''
        Locations in successive frames (ordered), starting at the give date.
        '''
        yield from self.__frame.locations_from(start)
        while True:
            start = self.next_frame(start)
            if start:
                yield from self.__frame.locations_from(start)
            else:
                return

    def at_location(self, date):
        return self.__frame.at_location(date)

    @classmethod
    def normalize(cls, spec):
        return str(Schedule(spec))

    @classmethod
    def include(cls, parent, child):
        '''
        Extend the parent (creating if necessary) to include the child.
        '''
        if parent is None:
            parent = Schedule('')

        def apply_with_none(f, a, b):
            if a:
                if b:
                    return f(a, b)
                else:
                    return a
            else:
                if b:
                    return b
                else:
                    return None

        parent.start = apply_with_none(min, parent.start, child.start)
        parent.finish = apply_with_none(max, parent.finish, child.finish)
        return parent


class DateOrdinals:

    def __init__(self, date):
        date = to_date(date)
        self.y = date.year - 1970
        self.m = 12 * self.y + date.month - 1
        self.d = (date - dt.date(1970, 1, 1)).days
        day = self.d + WEEK_OFFSET
        self.w = day // 7  # 1970-01-01 is Th
        self.dow = day % 7
        self.date = date
        self.ordinals = vars(self)

    def __str__(self):
        return format_date(self.date)


class Frame(ABC):

    def __init__(self, schedule):
        self.schedule = schedule

    def start_of_frame(self, date):
        '''
        The start date for the frame that includes or precedes the given date
        (ignoring range).
        '''
        date = DateOrdinals(date)
        n = (date.ordinals[self.schedule.frame_type] - self.schedule.offset) // self.schedule.repeat
        zero = add_date(self.zero, (self.schedule.offset, self.schedule.frame_type))
        return add_date(zero, (n * self.schedule.repeat, self.schedule.frame_type))

    def locations_from(self, start):
        '''
        Locations in a single frame, starting at the give date
        (so if the date isn't at the start of the frame you might not get all locations).
        '''
        start = to_date(start)
        frame = self.start_of_frame(start)
        ordinals = DateOrdinals(frame)
        for delta in self._location_offsets(ordinals.dow, self.length_in_days(start)):
            date = frame + dt.timedelta(days=delta)
            if self.schedule.in_range(date) and date >= start:
                yield date

    def _location_offsets(self, dow, limit):
        '''
        Generate offsets into frame for locations specified.
        Returns ordered values in [0, limit).
        '''
        # in future we could cache these for efficiency?
        # this should check against lower (0) and upper bounds (limit), but not range
        # or start value passed to dates()
        if self.schedule.locations:
            # locations may not be ordered, but is finite, so calculate a week in a batch
            # this guarantees that the offsets are ordered and unique (needed by dates())
            for week in range(1 + (dow + limit - 1) // 7):
                days = set()
                for location in self.schedule.locations:
                    if isinstance(location, int):
                        if 7*(week+1) >= dow + location > 7*week and location <= limit:
                            days.add(location - 1)
                    else:
                        n, day = location
                        if n and day < dow:
                            n += 1  # months only - numbering is not per week, but consecutive
                        if (n == 0 or week + 1 == n) and 0 <= week * 7 + day - dow < limit:
                            days.add(week*7 + day - dow)
                yield from sorted(days)
        else:
            yield from range(limit)

    def at_location(self, date):
        '''
        Does the given day coincide with the specification?
        '''
        try:
            date = to_date(date)
            return self.schedule.in_range(date) and date == next(self.locations_from(date))
        except StopIteration:
            return False

    @abstractmethod
    def length_in_days(self, date):
        '''
        The duration of a single frame containing the given date, in days.
        '''
        pass


class Day(Frame):

    zero = ZERO

    def length_in_days(self, date):
        return self.schedule.repeat


class Week(Frame):

    zero = ZERO - dt.timedelta(days=WEEK_OFFSET)

    def length_in_days(self, date):
        return self.schedule.repeat * 7


class Month(Frame):

    zero = ZERO

    def length_in_days(self, date):
        return monthrange(date.year, date.month)[1]


class Year(Frame):

    zero = ZERO

    def length_in_days(self, date):
        return (dt.date(date.year+1, 1, 1) - dt.date(date.year, 1, 1)).days

    def at_location(self, date):
        # locations not supported, so ignore for efficiency
        return self.schedule.in_range(date)




import datetime as dt
from logging import getLogger, basicConfig, DEBUG, StreamHandler, Formatter
from sys import stdout

import sqlalchemy as s
from sqlalchemy.orm import sessionmaker
from urwid import WidgetWrap, Pile, Edit, Filler

from ch2.squeal.binders import Binder
from ch2.squeal.support import Base
from ch2.squeal.types import Date
from ch2.uweird.tui.widgets import Integer

basicConfig()
log = getLogger()
log.setLevel(DEBUG)
handler = StreamHandler(stdout)
handler.setLevel(DEBUG)
formatter = Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')
handler.setFormatter(formatter)
log.addHandler(handler)


class Data(Base):

    __tablename__ = 'data'

    integer = s.Column(s.Integer, primary_key=True)
    text = s.Column(s.Text, nullable=False, server_default='')
    date = s.Column(Date)


class Database:

    def __init__(self):
        self._log = log
        self.engine = s.create_engine('sqlite:///:memory:', echo=True)
        self.__create_tables()
        self.session = sessionmaker(bind=self.engine)

    def __create_tables(self):
        self._log.info('Creating tables')
        Base.metadata.create_all(self.engine)


class DataWidget(WidgetWrap):

    def __init__(self):
        self.integer = Integer('An integer: ')
        self.text = Edit('Some text: ')
        super().__init__(Filler(Pile([self.integer, self.text]), valign='top'))


def test_bind():

    db = Database()
    session = db.session()

    data = Data(integer=42, text='abc')
    session.add(data)
    widget = DataWidget()
    binder = Binder(log, session, widget, Data, defaults={'integer': 42})

    assert binder.instance.text == 'abc', binder.instance.text
    assert widget.text.edit_text == 'abc', widget.text.edit_text
    assert binder.instance.integer == 42, binder.instance.integer
    assert widget.integer.state == 42, widget.integer.state

    # edit the text
    size = (10,)
    for key in ('right', 'right', 'delete'):
        widget.text.keypress(size, key)
    assert widget.text.edit_text == 'ab', widget.text.edit_text
    assert binder.instance.text == 'ab', binder.instance.text
    assert binder.instance.integer == 42, binder.instance.integer

    # modify the primary key
    size = (10, 10)
    for key in ('right', 'right'):
        widget.integer.keypress(size, key)
    try:
        widget.integer.keypress(size, 'delete')
        assert False, 'expected error because primary key'
    except:
        pass

    session.expunge_all()
    # no need to add data - it was saved when the first binder above did a query
    widget = DataWidget()
    binder = Binder(log, session, widget, Data, multirow=True, defaults={'integer': 42})

    assert binder.instance.text == 'abc', binder.instance.text
    assert widget.text.edit_text == 'abc', widget.text.edit_text
    assert binder.instance.integer == 42, binder.instance.integer
    assert widget.integer.state == 42, widget.integer.state

    # edit the text
    size = (10,)
    for key in ('home', 'right', 'right', 'delete'):
        log.info('Keypress %s' % key)
        widget.text.keypress(size, key)
    assert widget.text.edit_text == 'ab', widget.text.edit_text
    assert binder.instance.text == 'ab', binder.instance.text
    assert binder.instance.integer == 42, binder.instance.integer

    # modify the primary key
    size = (10, 10)
    for key in ('right', 'right', 'delete'):
        widget.integer.keypress(size, key)

    assert binder.instance.integer == 4, binder.instance.integer
    assert binder.instance.text == '', binder.instance.text

    # and back again
    widget.integer.keypress(size, '2')
    assert binder.instance.integer == 42, binder.instance.integer
    assert binder.instance.text == 'ab', binder.instance.text

    # try setting date
    binder.instance.date = dt.date(2018, 7, 1)
    session.commit()
    data = session.query(Data).filter(Data.integer == 42).one()
    assert data.date == dt.date(2018, 7, 1)




from .args import NO_DIARY
from ..config import default
from ..squeal.tables.activity import ActivityGroup
from ..squeal.tables.statistic import StatisticName


def default_config(args, log, db):
    '''
## default-config

    > ch2 default-config

Generate a simple initial configuration.

Please see the documentation at http://andrewcooke.github.io/choochoo
    '''
    no_diary = args[NO_DIARY]
    with db.session_context() as s:
        if not no_diary and s.query(StatisticName).count():
            raise Exception('The database already contains StatisticName entries')
        if s.query(ActivityGroup).count():
            raise Exception('The database already contains ActivityGroup entries')
    default(log, db, no_diary=no_diary)




import pandas as pd

from ..data.database import Data
from .args import ACTIVITY_GROUPS, SUB_COMMAND, ACTIVITY_JOURNALS, GROUP, START, FINISH, STATISTIC_NAMES, NAMES, \
    STATISTIC_JOURNALS, OWNER, CONSTRAINT, SCHEDULE, SOURCE_IDS, STATISTIC_QUARTILES, MONITOR_JOURNALS, PRINT, FORMAT, \
    CSV, DESCRIBE, MAX_COLUMNS, MAX_ROWS, WIDTH, MAX_COLWIDTH, SEGMENTS, SEGMENT_JOURNALS


def data(args, log, db):
    '''
## data

    > ch2 data COMMAND

Simple access to Pandas DataFrames - the same interface provided in Jupyter notebooks,
but accessed from the command line.

The format can be selected with `--print` (the default), `--csv` and `--describe`.

For full options see `ch2 data -h` and `ch2 data COMMAND -h`

### Examples

    > ch2 data --csv statistic-names

Will print details on all statistic names in CSV format.

    > ch2 data statistic-journals '%HR%' --constraint 'ActivityGroup "Bike"' --start 2018-01-01

Will print HR-related statistics from the start of 2018 for the given activity group.
    '''
    data = Data(log, db)

    if args[SUB_COMMAND] == ACTIVITY_GROUPS:
        frame = data.activity_groups()
    elif args[SUB_COMMAND] == ACTIVITY_JOURNALS:
        frame = data.activity_journals(args[GROUP], start=args[START], finish=args[FINISH])
    elif args[SUB_COMMAND] == SEGMENT_JOURNALS:
        frame = data.segment_journals()
    elif args[SUB_COMMAND] == SEGMENTS:
        frame = data.segments()
    elif args[SUB_COMMAND] == STATISTIC_NAMES:
        frame = data.statistic_names(*args[NAMES])
    elif args[SUB_COMMAND] == STATISTIC_JOURNALS:
        frame = data.statistic_journals(*args[NAMES], start=args[START], finish=args[FINISH],
                                        owner=args[OWNER], constraint=args[CONSTRAINT],
                                        schedule=args[SCHEDULE], source_ids=args[SOURCE_IDS])
    elif args[SUB_COMMAND] == STATISTIC_QUARTILES:
        frame = data.statistic_quartiles(*args[NAMES], start=args[START], finish=args[FINISH],
                                         owner=args[OWNER], constraint=args[CONSTRAINT],
                                         schedule=args[SCHEDULE], source_ids=args[SOURCE_IDS])
    elif args[SUB_COMMAND] == MONITOR_JOURNALS:
        frame = data.monitor_journals()
    else:
        raise Exception('Unexpected %s: %s' % (SUB_COMMAND, args[SUB_COMMAND]))

    pd.options.display.max_columns = args[MAX_COLUMNS]
    if args[MAX_COLWIDTH] is not None: pd.options.display.max_colwidth = args[MAX_COLWIDTH]
    pd.options.display.max_rows = args[MAX_ROWS]
    pd.options.display.width = args[WIDTH]

    if args[FORMAT] == PRINT:
        print(frame)
    elif args[FORMAT] == CSV:
        print(frame.to_csv())
    elif args[FORMAT] == DESCRIBE:
        print(frame.describe(include='all'))




from sqlalchemy.sql.functions import count

from . import WaypointCalculator
from ..names import SEGMENT_TIME, LATITUDE, LONGITUDE, DISTANCE, S, summaries, MIN, MSR, CNT, HEART_RATE, \
    SEGMENT_HEART_RATE, BPM, MAX
from ..waypoint import filter_none
from ...squeal.tables.activity import ActivityJournal
from ...squeal.tables.segment import SegmentJournal, Segment
from ...squeal.tables.statistic import StatisticName, StatisticJournal, StatisticJournalType


class SegmentStatistics(WaypointCalculator):

    def run(self, force=False, after=None):
        with self._db.session_context() as s:
            if 0 == s.query(count(Segment.id)).scalar():
                self._log.warn('No segments defined in database')
                return
        super().run(force=force, after=after)

    def _filter_statistic_journals(self, q):
        return q.filter(StatisticName.name == SEGMENT_TIME)

    def _filter_activity_journals(self, q):
        return q.join(SegmentJournal, SegmentJournal.activity_journal_id == ActivityJournal.id)

    def _names(self):
        return {LATITUDE: 'lat',
                LONGITUDE: 'lon',
                DISTANCE: 'distance',
                HEART_RATE: 'hr'}

    def _add_stats_from_waypoints(self, s, ajournal, waypoints):
        for sjournal in s.query(SegmentJournal). \
                filter(SegmentJournal.activity_journal == ajournal).all():
            StatisticJournal.add(self._log, s, SEGMENT_TIME, S, summaries(MIN, CNT, MSR), self,
                                 sjournal.segment, sjournal,
                                 (sjournal.finish - sjournal.start).total_seconds(),
                                 sjournal.start, StatisticJournalType.FLOAT)
            waypoints = [w for w in filter_none(self._names().values(), waypoints)
                         if sjournal.start <= w.time <= sjournal.finish]
            # weight by time gap so we don't bias towards more sampled times
            gaps = [(w1.time - w0.time, 0.5 * (w0.hr + w1.hr))
                    for w0, w1 in zip(waypoints, waypoints[1:])]
            if gaps:
                weighted = sum(dt.total_seconds() * hr for dt, hr in gaps)
                average = weighted / sum(dt.total_seconds() for dt, _ in gaps)
                StatisticJournal.add(self._log, s, SEGMENT_HEART_RATE, BPM, summaries(MAX, CNT, MSR), self,
                                     sjournal.segment, sjournal, average,
                                     sjournal.start, StatisticJournalType.FLOAT)
            else:
                self._log.warn('No Heart Rate data')

    def _delete_my_statistics(self, s, activity_group, after=None):
        s.commit()   # so that we don't have any risk of having something in the session that can be deleted
        statistic_name_ids = s.query(StatisticName.id). \
            filter(StatisticName.owner == self).cte()
        segment_journal_ids = s.query(SegmentJournal.id).join(Segment). \
            filter(Segment.activity_group == activity_group).cte()
        for repeat in range(2):
            if repeat:
                q = s.query(StatisticJournal)
            else:
                q = s.query(count(StatisticJournal.id))
            q = q.filter(StatisticJournal.statistic_name_id.in_(statistic_name_ids),
                         StatisticJournal.source_id.in_(segment_journal_ids))
            if after:
                q = q.filter(StatisticJournal.time >= after)
            self._log.debug(q)
            if repeat:
                q.delete(synchronize_session=False)
            else:
                n = q.scalar()
                if n:
                    self._log.warn('Deleting %d statistics for %s' % (n, activity_group))
                else:
                    self._log.warn('No statistics to delete for %s' % activity_group)
        s.commit()




from urwid import Text, emit_signal

from .focus import FocusWrap


class ImmutableStatefulText(FocusWrap):
    """
    A text class where the text depends on some state
    """

    def __init__(self, state):
        self._text = Text('')
        self._text._selectable = True
        super().__init__(self._text)
        self._state = state
        self._selectable = True
        self._update_text()

    @property
    def state(self):
        return self._state

    def _state_as_text(self):
        return str(self.state)

    def _update_text(self):
        self._text.set_text(self._state_as_text())
        self._invalidate()

    def keypress(self, size, key):
        # need this because selectable
        return key


class MutableStatefulText(ImmutableStatefulText):
    """
    A text class where the text depends on some state and
    the state may be changed
    """

    signals = ['change', 'postchange']

    def _get_state(self):
        return self._state

    def _set_state(self, state):
        self._set_state_external(state)

    def _set_state_external(self, state):
        self._set_state_internal(state)

    def _set_state_internal(self, state, signal=True):
        if state != self._state:
            old_state = self._state
            self._state = state
            # argument list to match Edit behaviour
            if signal:
                emit_signal(self, 'change', self, state)
            self._update_text()
            # this is similar behaviour to the edit widget
            if signal:
                emit_signal(self, 'postchange', self, old_state)

    state = property(_get_state, _set_state)




import datetime as dt
from calendar import month_name, day_abbr, Calendar, monthrange

from urwid import Columns, GridFlow, Pile, Text, Padding, emit_signal, connect_signal, WEIGHT

from .fixed import Fixed
from .focus import FocusFor, FocusAttr, FocusWrap, OnFocus
from .state import ImmutableStatefulText, MutableStatefulText
from ...lib.date import format_date

MONTHS = month_name
DAYS2 = list(map(lambda d: day_abbr[d][:2], Calendar(0).iterweekdays()))
DAYS3 = list(map(lambda d: day_abbr[d][:3], Calendar(0).iterweekdays()))


def clip_day(day, month, year): return min(day, monthrange(year, month)[1])


class DateKeyPressMixin:
    """
    Assumes self.state is a date.

    default should be 'd', 'w', 'm', 'y', 'D' or 'M' (capitals support alpha selection)
    delta should be +1 or -1
    """

    def __init__(self, default, delta=1):
        self.__is_alpha = default.isupper()
        self.__default = default.lower()
        self.__delta = delta

    def keypress(self, size, key):
        original_key = key
        delta = self.__delta
        if len(key) == 1 and key.isupper():
            key = key.lower()
            delta *= -1
        if self.__is_alpha and 'a' <= key <= 'z':
            return self.__alpha_keypress(delta, key, original_key)
        if self._command_map[key] == 'activate':
            key = self.__default
        if key == '=':
            self.state = dt.date.today()
            return
        if len(key) == 1:
            return self.__delta_keypress(delta, key, original_key)
        return original_key

    def __delta_keypress(self, delta, key, original_key):
        if key in '+- ':
            if key == '-': delta *= -1
            key = 'd' if self.__default == '=' else self.__default
        if '0' <= key <= '9':
            delta *= 10 if key == '0' else int(key)
            key = 'd' if self.__default == '=' else self.__default
        if key == 'w':
            key = 'd'
            delta *= 7
        if key == 'y':
            key = 'm'
            delta *= 12
        if key == 'd':
            self.state = self.state + dt.timedelta(days=delta)
            return
        elif key == 'm':
            year, month, day = self.state.year, self.state.month, self.state.day
            month += delta
            while month < 1:
                year -= 1
                month += 12
            while month > 12:
                year += 1
                month -= 12
            day = min(day, monthrange(year, month)[1])
            self.state = dt.date(year, month, day)
            return
        return original_key

    def __alpha_keypress(self, delta, key, original_key):
        is_month = self.__default == 'm'
        n = 12 if is_month else 7
        tries, date = 0, self.state
        while tries <= n:
            tries += 1
            if is_month:
                date = self.__add_month(date, delta)
                name = MONTHS[date.month]
            else:
                date += dt.timedelta(days=delta)
                name = DAYS2[date.weekday()]
            if name.lower().startswith(key):
                if is_month:
                    date = dt.date(date.year, date.month, clip_day(self.state.day, date.month, date.year))
                self.state = date
                return
        return original_key

    def __add_month(self, date, delta):
        year, month, day = date.year, date.month + delta, date.day
        while month < 1:
            year -= 1
            month += 12
        while month > 12:
            year += 1
            month -= 12
        day = clip_day(day, month, year)
        return dt.date(year, month, day)


class Month(DateKeyPressMixin, MutableStatefulText):

    def __init__(self, date, as_text=True):
        self._as_text = as_text
        MutableStatefulText.__init__(self, date)
        DateKeyPressMixin.__init__(self, 'M')

    def _state_as_text(self):
        if self._as_text:
            return MONTHS[self.state.month]
        else:
            return '%02d' % self.state.month


class MonthBar(OnFocus):

    def __init__(self, date, as_text=True, bar=None):
        super().__init__(Month(date, as_text=as_text),
                         '+/- to inc/dec; first letter to advance; d/m/y/D/M/Y/= as elsewhere', bar)
        self.widget = self._w


class Year(DateKeyPressMixin, MutableStatefulText):

    def __init__(self, date):
        MutableStatefulText.__init__(self, date)
        DateKeyPressMixin.__init__(self, 'y')

    def _state_as_text(self):
        return str(self.state.year)


class YearBar(OnFocus):

    def __init__(self, date, bar=None):
        super().__init__(Year(date), '+/- to inc/dec; digit to jump to nearest; d/m/y/D/M/Y/= as elsewhere', bar)
        self.widget = self._w


class Day(ImmutableStatefulText):

    signals = ['click']

    def _state_as_text(self):
        return '%2s' % self.state.day

    def keypress(self, size, key):
        if self._command_map[key] == 'activate':
            emit_signal(self, 'click', self, self.state)
        else:
            return key

    def mouse_event(self, size, event, button, col, row, focus):
        if event == 'mouse release' and focus:
            emit_signal(self, 'click', self, self.state)
            return True
        else:
            return False


class Days(FocusWrap):

    def __init__(self, date, calendar):
        self._date = date
        super().__init__(self._make(calendar))

    def _make(self, calendar):
        # if we do this as single gridflow then focus doesn't move down into dates
        # so instead put names in pile
        names = GridFlow(list(map(lambda d: Text(('plain', d)), DAYS2)), 2, 1, 0, 'left')
        prev = dt.date(self._date.year if self._date.month > 1 else self._date.year - 1,
                       self._date.month - 1 if self._date.month > 1 else 12, 1)
        next = dt.date(self._date.year if self._date.month < 12 else self._date.year + 1,
                       self._date.month + 1 if self._date.month < 12 else 1, 1)
        prev_days = monthrange(prev.year, prev.month)[1]
        curr_days = monthrange(self._date.year, self._date.month)[1]
        first_day = dt.date(self._date.year, self._date.month, 1).weekday()  # mon 0
        total_days = first_day + curr_days
        extra_days = 7 - total_days % 7
        if extra_days == 7:
            extra_days = 0
        else:
            total_days += extra_days
        dates = [FocusAttr(Day(dt.date(prev.year, prev.month, i)), 'unimportant')
                 for i in range(prev_days - first_day + 1, prev_days + 1)]
        dates.extend([FocusAttr(Day(dt.date(self._date.year, self._date.month, i)),
                                plain='selected' if i == self._date.day else 'plain',
                                focus='selected-focus' if i == self._date.day else 'plain-focus')
                      for i in range(1, curr_days + 1)])
        dates.extend([FocusAttr(Day(dt.date(next.year, next.month, i)), 'unimportant')
                      for i in range(1, extra_days + 1)])
        for day in dates:
            connect_signal(day._original_widget, 'click', calendar.date_change)
        dates = GridFlow(dates, 2, 1, 0, 'left')
        return Pile([names, dates])


class StatefulSymbol(MutableStatefulText):

    signals = ['click']

    def __init__(self, state, symbol):
        self._symbol = symbol
        # call super after saving symbol so we can display it
        super().__init__(state)

    def _state_as_text(self):
        return self._symbol

    def mouse_event(self, size, event, button, col, row, focus):
        if event == 'mouse release' and focus:
            self.keypress(size, ' ')
            return True
        else:
            return False


class QuickChange(DateKeyPressMixin, StatefulSymbol):

    def __init__(self, date, symbol, sign):
        StatefulSymbol.__init__(self, date, symbol)
        DateKeyPressMixin.__init__(self, 'd', sign)


DKMSG = {(-1, 'd'): '+/spc/enter/d to dec day; -/D to inc; m/y to dec month/year; M/Y to inc; = now',
         ( 1, 'd'): '+/spc/enter/d to inc day; -/D to dec; m/y to inc month/year; M/Y to dec; = now',
         (-1, 'm'): '+/spc/enter/m to dec month; -/M to inc; d/y to dec day/year; D/Y to inc; = now',
         ( 1, 'm'): '+/spc/enter/m to inc month; -/M to dec; d/y to inc day/year; D/Y to dec; = now',
         (-1, 'y'): '+/spc/enter/y to dec year; -/Y to inc; d/m to dec day/month; D/M to inc; = now',
         ( 1, 'y'): '+/spc/enter/y to inc year; -/Y to dec; d/m to inc day/month; D/M to dec; = now'}


class QuickChangeBar(OnFocus):

    def __init__(self, date, symbol, sign, bar=None):
        super().__init__(QuickChange(date, symbol, sign), DKMSG[(sign, 'd')], bar)
        self.widget = self._w


class Today(DateKeyPressMixin, StatefulSymbol):

    def __init__(self, date, symbol):
        StatefulSymbol.__init__(self, date, symbol)
        DateKeyPressMixin.__init__(self, '=')


class TodayBar(OnFocus):

    def __init__(self, date, symbol, bar=None):
        super().__init__(Today(date, symbol),
                         '%s/spc/enter now; +/d/m/y to inc day/month/year; -/D/M/Y to dec' % symbol, bar)
        self.widget = self._w


class BaseDate(FocusWrap):

    signals = ['change', 'postchange']

    def __init__(self, log, bar=None, date=None):
        self._log = log
        if not date: date = dt.date.today()
        self._date = date
        self._bar = bar
        super().__init__(self._make())

    def __get_state(self):
        return self._date

    def __set_state(self, date):
        self.date_change(None, date)

    state = property(__get_state, __set_state)

    def _make(self):
        raise NotImplemented()

    def date_change(self, unused_widget, date):
        if date != self._date:
            self._log.info('Date has changed: %s - %s' % (format_date(self._date), format_date(date)))
            old_date = date
            self._date = date
            self._log.debug('Sending change signal for date change')
            # again, arg convention matches Edit
            emit_signal(self, 'change', self, self._date)
            focus = FocusFor(self._w, self._log)
            self._w = self._make()
            focus.to(self._w)
            self._log.debug('Sending change signal for date postchange')
            emit_signal(self, 'postchange', self, old_date)


class Calendar(BaseDate):
    """
    Displays a text calendar with signal when date changed.
    """

    def _make(self):
        down = QuickChangeBar(self._date, '<', -1, bar=self._bar)
        connect_signal(down.widget, 'change', self.date_change)
        today = TodayBar(self._date, '=', bar=self._bar)
        connect_signal(today.widget, 'change', self.date_change)
        up = QuickChangeBar(self._date, '>', 1, bar=self._bar)
        connect_signal(up.widget, 'change', self.date_change)
        month = MonthBar(self._date, bar=self._bar)
        connect_signal(month.widget, 'change', self.date_change)
        year = YearBar(self._date, bar=self._bar)
        connect_signal(year.widget, 'change', self.date_change)
        title = Columns([(1, FocusAttr(down)),
                         (1, FocusAttr(today)),
                         (1, FocusAttr(up)),
                         (WEIGHT, 1, Padding(FocusAttr(month), align='center', width='pack')),
                         (4, FocusAttr(year)),
                         ])
        return Fixed(Pile([title, Days(self._date, self)]), 20)


class DayOfMonth(DateKeyPressMixin, MutableStatefulText):

    def __init__(self, state):
        MutableStatefulText.__init__(self, state)
        DateKeyPressMixin.__init__(self, 'd')

    def _state_as_text(self):
        return '%02d' % self.state.day


class DayOfMonthBar(OnFocus):

    def __init__(self, state, bar=None):
        super().__init__(DayOfMonth(state),
                         '+/spc/enter/d to inc day; -/D to dec; m/y to inc month/year; M/Y to dec; = now', bar)
        self.widget = self._w


class DayOfWeek(DateKeyPressMixin, MutableStatefulText):

    def __init__(self, state):
        MutableStatefulText.__init__(self, state)
        DateKeyPressMixin.__init__(self, 'D')

    def _state_as_text(self):
        return DAYS3[self.state.weekday()]


class DayOfWeekBar(OnFocus):

    def __init__(self, state, bar=None):
        super().__init__(DayOfWeek(state),
                         '+/spc/enter/d to inc day; first letter to advance; = now', bar)
        self.widget = self._w


class TextDate(BaseDate):

    def _make(self):
        down = QuickChangeBar(self._date, '<', -1, bar=self._bar)
        connect_signal(down.widget, 'change', self.date_change)
        up = QuickChangeBar(self._date, '>', 1, bar=self._bar)
        connect_signal(up.widget, 'change', self.date_change)
        year = YearBar(self._date, bar=self._bar)
        connect_signal(year.widget, 'change', self.date_change)
        month = MonthBar(self._date, as_text=False, bar=self._bar)
        connect_signal(month.widget, 'change', self.date_change)
        day_of_month = DayOfMonthBar(self._date, bar=self._bar)
        connect_signal(day_of_month.widget, 'change', self.date_change)
        day_of_week = DayOfWeekBar(self._date, bar=self._bar)
        connect_signal(day_of_week.widget, 'change', self.date_change)
        return Columns([(1, FocusAttr(down)),
                        (1, Text(" ")),
                        (4, Padding(FocusAttr(year), align='center', width='pack')),
                        (1, Text("-")),
                        (2, Padding(FocusAttr(month), align='center', width='pack')),
                        (1, Text("-")),
                        (2, Padding(FocusAttr(day_of_month), align='center', width='pack')),
                        (1, Text(" ")),
                        (3, Padding(FocusAttr(day_of_week), align='center', width='pack')),
                        (1, Text(" ")),
                        (1, FocusAttr(up)),
                        ])




from contextlib import contextmanager

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.functions import count

from .support import Base
from .tables.activity import ActivityGroup, ActivityTimespan, ActivityJournal
from .tables.constant import SystemConstant, Constant
from .tables.monitor import MonitorJournal
from .tables.pipeline import Pipeline
from .tables.segment import Segment, SegmentJournal
from .tables.source import Source, Interval
from .tables.statistic import StatisticName, StatisticJournalFloat, StatisticJournalText, StatisticJournalInteger, \
    StatisticJournal
from .tables.topic import TopicJournal, Topic
from ..command.args import DATABASE, NamespaceWithVariables, NO_OP, parser
from ..lib.log import make_log

# import these so they are "created"
Source,  Interval,
ActivityGroup, ActivityJournal, ActivityTimespan,
Topic, TopicJournal,
StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat, StatisticJournalText,
Segment, SegmentJournal
Pipeline
MonitorJournal
Constant, SystemConstant


# https://stackoverflow.com/questions/13712381/how-to-turn-on-pragma-foreign-keys-on-in-sqlalchemy-migration-script-or-conf
@event.listens_for(Engine, "connect")
def fk_pragma_on_connect(dbapi_con, _con_record):
    cursor = dbapi_con.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


class Database:

    def __init__(self, args, log):
        self._log = log
        path = args.file(DATABASE)
        self._log.info('Using database at %s' % path)
        self.engine = create_engine('sqlite:///%s' % path, echo=False)
        self.session = sessionmaker(bind=self.engine)
        self.__create_tables()

    def __create_tables(self):
        if self.is_empty(tables=True):
            self._log.info('Creating tables')
            Base.metadata.create_all(self.engine)

    @contextmanager
    def session_context(self):
        session = self.session()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def is_empty(self, tables=False):
        if tables:
            # https://stackoverflow.com/questions/33053241/sqlalchemy-if-table-does-not-exist
            return not self.engine.dialect.has_table(self.engine, Source.__tablename__)
        else:
            with self.session_context() as s:
                n_topics = s.query(count(Topic.id)).scalar()
                n_activities = s.query(count(ActivityGroup.id)).scalar()
                n_statistics = s.query(count(StatisticName.id)).scalar()
            return not (n_topics + n_activities + n_statistics)


def add(s, instance):
    s.add(instance)
    return instance


def connect(args):
    '''
    Bootstrap from commandline-like args.
    '''
    p = parser()
    if len(args) == 1:
        args = args[0].split()
    elif args:
        args = list(args)
    else:
        args = []
    args.append(NO_OP)
    ns = NamespaceWithVariables(p.parse_args(args))
    log = make_log(ns)
    db = Database(ns, log)
    return ns, log, db





import datetime as dt
from abc import ABC, abstractmethod
from json import dumps, loads

from sqlalchemy import Column, Integer, ForeignKey, Text, Boolean, desc
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound

from .source import Source, SourceType
from .statistic import STATISTIC_JOURNAL_CLASSES, StatisticJournal
from ..support import Base
from ..types import Cls, Json, lookup_cls, short_cls
from ...lib.date import local_date_to_time, format_time


class Constant(Source):

    __tablename__ = 'constant'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    statistic_journal_type = Column(Integer, nullable=False)
    # this could be the statistic_name or it could contain more info related to constraint
    name = Column(Text, nullable=False, index=True)
    statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'), nullable=False)
    statistic_name = relationship('StatisticName')
    single = Column(Boolean, nullable=False, server_default='0')
    validate_cls = Column(Cls)
    validate_args = Column(Json, nullable=False, server_default=dumps(()))
    validate_kargs = Column(Json, nullable=False, server_default=dumps({}))

    def validate(self, sjournal):
        if self.validate_cls:
            if self.validate_args is None or self.validate_kargs is None:
                raise Exception('Missing args or kargs for %s' % self)
            validate = self.validate_cls(*self.validate_args, **self.validate_kargs)
            validate.validate(self, sjournal)

    def add_value(self, s, value, time=None, date=None):
        from ..database import add
        if time and date:
            raise Exception('Specify one or none of time and date for %s' % self)
        if not time and not date:
            time = 0.0  # important this is a float and not an int (or it would be an erroneous date)
        if date:
            time = local_date_to_time(date)
        if time and self.single:
            raise Exception('%s was given time %s but is not time-variable' % (self, format_time(time)))
        sjournal = STATISTIC_JOURNAL_CLASSES[self.statistic_journal_type](
            statistic_name=self.statistic_name, source=self, value=value, time=time)
        if self.validate_cls:
            self.validate(sjournal)
        return add(s, sjournal)

    def at(self, s, time=None, date=None):
        if time and date:
            raise Exception('Specify one or none of time and date for %s' % self)
        if date:
            time = local_date_to_time(date)
        if not time:
            time = dt.datetime.now()
        return s.query(StatisticJournal). \
            filter(StatisticJournal.statistic_name == self.statistic_name,
                   StatisticJournal.time <= time). \
            order_by(desc(StatisticJournal.time)).limit(1).one_or_none()

    @classmethod
    def get(cls, s, name):
        try:
            return s.query(Constant).filter(Constant.name == name).one()
        except NoResultFound:
            raise Exception('Could not find Constant for %s' % name)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.CONSTANT
    }

    def __str__(self):
        return 'Constant "%s"' % self.name


class SystemConstant(Base):

    __tablename__ = 'system_constant'

    name = Column(Text, primary_key=True)
    value = Column(Text, nullable=False)

    TIMEZONE = 'timezone'
    LOCK = 'lock'

    @classmethod
    def is_locked(cls, s):
        s.commit()
        return s.query(SystemConstant). \
            filter(SystemConstant.name == SystemConstant.LOCK).one_or_none()

    @classmethod
    def assert_unlocked(cls, s, allow=None):
        s.commit()
        lock = cls.is_locked(s)
        if lock and (not allow or lock.value != short_cls(allow)):
            raise Exception('Database is locked.  See `ch2 help unlock`.')

    @classmethod
    def acquire_lock(cls, s, owner):
        s.commit()
        cls.assert_unlocked(s, allow=owner)
        s.add(SystemConstant(name=SystemConstant.LOCK, value=short_cls(owner)))
        s.commit()

    @classmethod
    def release_lock(cls, s, owner):
        s.commit()
        s.query(SystemConstant). \
            filter(SystemConstant.name == SystemConstant.LOCK,
                   SystemConstant.value == short_cls(owner)).delete()

    @classmethod
    def reset_lock(cls, s):
        s.commit()
        s.query(SystemConstant). \
            filter(SystemConstant.name == SystemConstant.LOCK).delete()


class Validate(ABC):

    @abstractmethod
    def validate(self, constant, sjournal):
        raise NotImplementedError()


class ValidateError(Exception): pass


class ValidateNamedTuple(Validate):

    def __init__(self, tuple_cls=None):
        if not tuple_cls:
            raise Exception('Add tuple_cls to validator_kargs')
        self.__tuple_cls = lookup_cls(tuple_cls)

    def validate(self, constant, sjournal):
        value = None
        try:
            value = loads(sjournal.value)
        except Exception as e:
            raise ValidateError('Could not unpack JSON value for %s from "%s": %s' %
                                (sjournal.value, constant.name, e))
        try:
            self.__tuple_cls(**value)
        except Exception as e:
            raise ValidateError('Could not create %s from "%s" for "%s": %s' %
                                (self.__tuple_cls, value, constant.name, e))






from sqlalchemy import Column, Integer, ForeignKey, Float, Text, UniqueConstraint
from sqlalchemy.orm import relationship

from .source import SourceType, Source
from ..support import Base
from ..types import Time
from ...lib.date import format_time


class SegmentJournal(Source):

    __tablename__ = 'segment_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    segment_id = Column(Integer, ForeignKey('segment.id', ondelete='cascade'),
                        nullable=False, index=True)
    segment = relationship('Segment')
    activity_journal_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal', foreign_keys=[activity_journal_id])
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(segment_id, activity_journal_id, start, finish)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SEGMENT
    }

    def __str__(self):
        return 'Segment Journal %s to %s' % (format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish


class Segment(Base):

    __tablename__ = 'segment'

    id = Column(Integer, primary_key=True)
    # need this because (1) segments do depend on activity group and (2) we want separate stats
    # depending on segment/group so need this as a constraint.
    activity_group_id = Column(Integer, ForeignKey('activity_group.id'), nullable=False)
    activity_group = relationship('ActivityGroup')
    start_lat = Column(Float, nullable=False)
    start_lon = Column(Float, nullable=False)
    finish_lat = Column(Float, nullable=False)
    finish_lon = Column(Float, nullable=False)
    distance = Column(Float, nullable=False)
    name = Column(Text, nullable=False, index=True)
    description = Column(Text)

    @property
    def start(self):
        return self.start_lon, self.start_lat

    @start.setter
    def start(self, lon_lat):
        self.start_lon, self.start_lat = lon_lat

    @property
    def finish(self):
        return self.finish_lon, self.finish_lat

    @finish.setter
    def finish(self, lon_lat):
        self.finish_lon, self.finish_lat = lon_lat

    def coords(self, start):
        if start:
            return self.start
        else:
            return self.finish

    def __str__(self):
        return 'Segment "%s/%s"' % (self.name, self.activity_group.name)




from abc import abstractmethod

from sqlalchemy.sql.functions import count

from .. import DbPipeline
from ..waypoint import WaypointReader
from ...lib.schedule import Schedule
from ...squeal.tables.activity import ActivityJournal, ActivityGroup
from ...squeal.tables.pipeline import Pipeline
from ...squeal.tables.source import Interval
from ...squeal.tables.statistic import StatisticJournal, StatisticName, StatisticJournalFloat
from ...squeal.types import short_cls


def run_pipeline_after(log, db, type, after=None, force=False, like=None):
    with db.session_context() as s:
        for cls, args, kargs in Pipeline.all(log, s, type, like=like):
            log.info('Running %s (%s, %s)' % (short_cls(cls), args, kargs))
            cls(log, db, *args, **kargs).run(force=force, after=after)


def run_pipeline_paths(log, db, type, paths, force=False, like=None):
    with db.session_context() as s:
        for cls, args, kargs in Pipeline.all(log, s, type, like=like):
            log.info('Running %s (%s, %s)' % (short_cls(cls), args, kargs))
            cls(log, db, *args, **kargs).run(paths, force=force)


class IntervalCalculator(DbPipeline):
    '''
    Support for calculations associated with intervals.
    '''

    def run(self, force=False, after=None):
        schedule = Schedule(self._assert_karg('schedule'))
        self.run_schedule(force=force, after=after, schedule=schedule)

    def run_schedule(self, force=False, after=None, schedule=None):
        if force:
            self._delete(after=after)
        self._run_calculations(schedule)

    @abstractmethod
    def _run_calculations(self, schedule):
        raise NotImplementedError()

    def _delete(self, after=None):
        self._delete_intervals(after)

    def _delete_intervals(self, after=None):
        # we delete the intervals that all summary statistics depend on and they will cascade
        with self._db.session_context() as s:
            for repeat in range(2):
                if repeat:
                    q = s.query(Interval)
                else:
                    q = s.query(count(Interval.id))
                q = self._filter_intervals(q)
                if after:
                    q = q.filter(Interval.finish > after)
                if repeat:
                    for interval in q.all():
                        self._log.debug('Deleting %s' % interval)
                        s.delete(interval)
                else:
                    n = q.scalar()
                    if n:
                        self._log.warn('Deleting %d intervals' % n)
                    else:
                        self._log.warn('No intervals to delete')

    def _filter_intervals(self, q):
        return q.filter(Interval.owner == self)


class ActivityCalculator(DbPipeline):
    '''
    Support for calculations associated with activity journals.
    '''

    def run(self, force=False, after=None):
        with self._db.session_context() as s:
            for activity_group in s.query(ActivityGroup).all():
                self._log.debug('Checking statistics for activity %s' % activity_group.name)
                if force:
                    self._delete_my_statistics(s, activity_group, after=after)
                self._run_activity(s, activity_group)

    def _run_activity(self, s, activity_group):
        # which activity journals don't have data?
        q1 = s.query(StatisticJournal.source_id). \
            join(StatisticName). \
            filter(StatisticName.owner == self,
                   StatisticName.constraint == activity_group)
        q1 = self._filter_statistic_journals(q1)
        statistics = q1.cte()
        q2 = s.query(ActivityJournal)
        q2 = self._filter_activity_journals(q2)
        q2 = q2.outerjoin(statistics). \
            filter(ActivityJournal.activity_group == activity_group,
                   statistics.c.source_id == None)
        for ajournal in q2.order_by(ActivityJournal.start).all():
            self._log.info('Running %s for %s' % (short_cls(self), ajournal))
            self._add_stats(s, ajournal)

    @abstractmethod
    def _filter_statistic_journals(self, q):
        raise NotImplementedError()

    def _filter_activity_journals(self, q):
        return q

    @abstractmethod
    def _add_stats(self, s, ajournal):
        raise NotImplementedError()

    def _add_float_stat(self, s, ajournal, name, summary, value, units, time=None):
        if time is None:
            time = ajournal.start
        StatisticJournalFloat.add(self._log, s, name, units, summary, self,
                                  ajournal.activity_group, ajournal, value, time)

    def _delete_my_statistics(self, s, activity_group, after=None):
        '''
        Delete all statistics owned by this class and in the activity group.
        Fast because in-SQL.
        '''
        s.commit()   # so that we don't have any risk of having something in the session that can be deleted
        statistic_name_ids = s.query(StatisticName.id). \
            filter(StatisticName.owner == self).cte()
        activity_journal_ids = s.query(ActivityJournal.id). \
            filter(ActivityJournal.activity_group == activity_group).cte()
        for repeat in range(2):
            if repeat:
                q = s.query(StatisticJournal)
            else:
                q = s.query(count(StatisticJournal.id))
            q = q.filter(StatisticJournal.statistic_name_id.in_(statistic_name_ids),
                         StatisticJournal.source_id.in_(activity_journal_ids))
            if after:
                q = q.filter(StatisticJournal.time >= after)
            self._log.debug(q)
            if repeat:
                q.delete(synchronize_session=False)
            else:
                n = q.scalar()
                if n:
                    self._log.warn('Deleting %d statistics for %s' % (n, activity_group))
                else:
                    self._log.warn('No statistics to delete for %s' % activity_group)
        s.commit()


class WaypointCalculator(ActivityCalculator):

    def _add_stats(self, s, ajournal):
        owner = self._assert_karg('owner')
        waypoints = list(WaypointReader(self._log).read(s, ajournal, self._names(), owner))
        if waypoints:
            self._add_stats_from_waypoints(s, ajournal, waypoints)
        else:
            self._log.warn('No statistics for %s' % ajournal)

    @abstractmethod
    def _add_stats_from_waypoints(self, s, ajournal, waypoints):
        raise NotImplementedError()

    @abstractmethod
    def _names(self):
        raise NotImplementedError()





import datetime as dt
from abc import abstractmethod
from collections import namedtuple
from re import compile
from struct import unpack

from .support import Named, Rows
from ...lib.data import WarnDict, WarnList

LITTLE, BIG = 0, 1


class AbstractType(Named):

    def __init__(self, log, name, size, base_type=None):
        super().__init__(log, name)
        self.base_type = base_type
        self.size = size

    @abstractmethod
    def profile_to_internal(self, cell_contents):
        raise NotImplementedError('%s: %s' % (self.__class__.__name__, self.name))

    @abstractmethod
    def parse(self, bytes, count, endian, timestamp, **options):
        raise NotImplementedError('%s: %s' % (self.__class__.__name__, self.name))


class BaseType(AbstractType):

    def __init__(self, log, name, size, func):
        super().__init__(log, name, size)
        self.__func = func

    def profile_to_internal(self, cell_contents):
        return self.__func(cell_contents)


class StructSupport(BaseType):

    def _pack_bad(self, value):
        bad = (bytearray(self.size), bytearray(self.size))
        for endian in (LITTLE, BIG):
            bytes = value
            for i in range(self.size):
                j = i if endian == LITTLE else self.size - i - 1
                bad[endian][j] = bytes & 0xff
                bytes >>= 8
        return bad

    def _is_bad(self, data, bad):
        size = len(bad)
        count = len(data) // size
        return all(bad == data[size*i:size*(i+1)] for i in range(count))

    def _unpack(self, data, formats, bad, count, endian):
        if self._is_bad(data, bad[endian]):
            return None
        else:
            return unpack(formats[endian] % count, data[0:count * self.size])


class String(BaseType):

    # it may seem weird this returns a singleton tuple.  shouldn't it be treated like a
    # character field where count is the mutliplicity?  but dynamic fields treat it as
    # a single value...

    def __init__(self, log, name):
        super().__init__(log, name, 1, str)

    def parse(self, bytes, count, endian, timestamp, **options):
        return (str(b''.join(byte for byte in unpack('%dc' % count, bytes) if byte != b'\0'),
                    encoding='utf-8'),)


class Boolean(BaseType):

    def __init__(self, log, name):
        super().__init__(log, name, 1, bool)

    def parse(self, bytes, count, endian, timestamp, **options):
        return tuple(bool(byte) for byte in bytes)


class AutoInteger(StructSupport):

    pattern = compile(r'^([su]?)int(\d{1,2})(z?)$')

    size_to_format = {1: 'b', 2: 'h', 4: 'i', 8: 'q'}

    def __init__(self, log, name):
        match = self.pattern.match(name)
        self.signed = match.group(1) != 'u'
        bits = int(match.group(2))
        if bits % 8:
            raise Exception('Size of %r not a multiple of 8 bits' % name)
        super().__init__(log, name, bits // 8, self.int)
        if self.size not in self.size_to_format:
            raise Exception('Cannot unpack %d bytes as an integer' % self.size)
        format = self.size_to_format[self.size]
        if not self.signed:
            format = format.upper()
        self.formats = ['<%d' + format, '>%d' + format]
        self.bad = self._pack_bad(0 if match.group(3) == 'z' else 2 ** (bits - (1 if self.signed else 0)) - 1)

    @staticmethod
    def int(cell):
        if isinstance(cell, int):
            return cell
        else:
            return int(cell, 0)

    def parse(self, data, count, endian, timestamp, **options):
        result = self._unpack(data, self.formats, self.bad, count, endian)
        if result is not None and self.size == 1:
            result = bytes(result)
        return result


class AliasInteger(AutoInteger):

    def __init__(self, log, name, spec):
        super().__init__(log, spec)
        self.name = name


def timestamp_to_time(timestamp, tzinfo=dt.timezone.utc):
    return dt.datetime(1989, 12, 31, tzinfo=tzinfo) + dt.timedelta(seconds=timestamp)


def time_to_timestamp(time, tzinfo=dt.timezone.utc):
    return int((time - dt.datetime(1989, 12, 31, tzinfo=tzinfo)).total_seconds())


class Date(AliasInteger):

    def __init__(self, log, name, utc=True):
        super().__init__(log, name, 'uint32')
        self.__tzinfo = dt.timezone.utc if utc else None

    @staticmethod
    def convert(time, tzinfo=dt.timezone.utc):
        if time is not None and time >= 0x10000000:
            return timestamp_to_time(time, tzinfo=tzinfo)
        else:
            return time

    def parse(self, data, count, endian, timestamp, raw_time=False, **options):
        times = super().parse(data, count, endian, timestamp, raw_time=raw_time, **options)
        if not raw_time:
            times = tuple(self.convert(time, tzinfo=self.__tzinfo) for time in times)
        return times


class Date16(AliasInteger):

    def __init__(self, log, name, utc=True):
        super().__init__(log, name, 'uint16')
        self.__tzinfo = dt.timezone.utc if utc else None

    @staticmethod
    def convert(time, timestamp, tzinfo=dt.timezone.utc):
        # https://www.thisisant.com/forum/viewthread/6374
        current = time_to_timestamp(timestamp, tzinfo=tzinfo)
        current += (time - (current & 0xffff)) & 0xffff
        return timestamp_to_time(current, tzinfo=tzinfo)

    def parse(self, data, count, endian, timestamp, raw_time=False, **options):
        times = super().parse(data, count, endian, timestamp, raw_time=raw_time, **options)
        if not raw_time:
            times = tuple(self.convert(time, timestamp, tzinfo=self.__tzinfo) for time in times)
        return times


class AutoFloat(StructSupport):

    pattern = compile(r'^float(\d{1,2})$')

    size_to_format = {2: 'e', 4: 'f', 8: 'd'}

    def __init__(self, log, name):
        match = self.pattern.match(name)
        bits = int(match.group(1))
        if bits % 8:
            raise Exception('Size of %r not a multiple of 8 bits' % name)
        super().__init__(log, name, bits // 8, float)
        if self.size not in self.size_to_format:
            raise Exception('Cannot unpack %d bytes as a float' % self.size)
        format = self.size_to_format[self.size]
        self.formats = ['<%d' + format, '>%d' + format]
        self.bad = self._pack_bad(2 ** bits - 1)

    def parse(self, data, count, endian, timestamp, **options):
        return self._unpack(data, self.formats, self.bad, count, endian)


class Mapping(AbstractType):

    def __init__(self, log, row, rows, types, warn=False):
        name = row.type_name
        base_type_name = row.base_type
        base_type = types.profile_to_type(base_type_name, auto_create=True)
        super().__init__(log, name, base_type.size, base_type=base_type)
        self._profile_to_internal = WarnDict(log, 'No internal value for profile %r') if warn else dict()
        self._internal_to_profile = WarnDict(log, 'No profile value for internal %r') if warn else dict()
        while rows:
            peek = rows.peek()
            if peek.type_name or peek.value_name is None or peek.value is None:
                return
            self.__add_mapping(next(rows))

    def profile_to_internal(self, cell_contents):
        return self._profile_to_internal[cell_contents]

    def internal_to_profile(self, value):
        return self._internal_to_profile[value]

    def safe_internal_to_profile(self, value):
        try:
            return self.internal_to_profile(value)
        except KeyError:
            return value

    def parse(self, bytes, size, endian, timestamp, map_values=True, **options):
        values = self.base_type.parse(bytes, size, endian, timestamp, map_values=map_values, **options)
        if map_values and values:
            values = tuple(self.safe_internal_to_profile(value) for value in values)
        return values

    def __add_mapping(self, row):
        profile = row.value_name
        internal = self.base_type.profile_to_internal(row.value)
        self._profile_to_internal[profile] = internal
        self._internal_to_profile[internal] = profile


BASE_TYPE_NAMES = ['enum', 'sint8', 'uint8', 'sint16', 'uint16', 'sint32', 'uint32',
                   'string', 'float32', 'float64',
                   'uint8z', 'uint16z', 'uint32z', 'byte', 'sint64', 'uint64', 'uint64z']


class Types:

    def __init__(self, log, sheet, warn=False):
        self.__log = log
        self.__profile_to_type = WarnDict(log, 'No type for profile %r')
        # these are not 'base types' in the same sense as types having base types.
        # rather, they are the 'base (integer) types' described in the docs
        self.base_types = WarnList(log, 'No base type for number %r')
        self.__add_known_types()
        rows = Rows(sheet, wrapper=Row)
        for row in rows:
            if row[0] and row.type_name[0].isupper():
                self.__log.debug('Skipping %s' % (row,))
            elif row[0]:
                # self.__log.info('Parsing type %s' % row[0])
                self.__add_type(Mapping(self.__log, row, rows, self, warn=warn))

    def __add_known_types(self):
        # these cannot be inferred from name
        self.__add_type(String(self.__log, 'string'))
        self.__add_type(AliasInteger(self.__log, 'enum', 'uint8'))
        self.__add_type(AliasInteger(self.__log, 'byte', 'uint8'))
        # these can be inferred
        for name in BASE_TYPE_NAMES:
            self.profile_to_type(name, auto_create=True)
            self.base_types.append(self.profile_to_type(name))
        # this is in the spreadsheet, but not in the doc
        self.__add_type(Boolean(self.__log, 'bool'))
        # these are defined in the spreadsheet, but the interpretation is in comments
        self.__add_type(Date(self.__log, 'date_time', utc=True))
        self.__add_type(Date(self.__log, 'local_date_time', utc=False))
        self.__add_type(Date16(self.__log, 'timestamp_16', utc=True))

    def __add_type(self, type):
        if type.name in self.__profile_to_type:
            duplicate = self.__profile_to_type[type.name]
            if duplicate.size == type.size:
                self.__log.warn('Ignoring duplicate type for %r' % type.name)
            else:
                raise Exception('Duplicate type for %r with differing size (%d  %d)' %
                                (type.name, type.size, duplicate.size))
        else:
            self.__profile_to_type[type.name] = type

    def is_type(self, name):
        return name in self.__profile_to_type

    def profile_to_type(self, name, auto_create=False):
        try:
            return self.__profile_to_type[name]
        except KeyError:
            if auto_create:
                for cls in (AutoFloat, AutoInteger):
                    match = cls.pattern.match(name)
                    if match:
                        self.__log.warn('Auto-adding type %s for %r' % (cls.__name__, name))
                        self.__add_type(cls(self.__log, name))
                        return self.profile_to_type(name)
            raise


class Row(namedtuple('BaseRow',
                     'type_name, base_type, value_name, value, comment')):

    __slots__ = ()

    def __new__(cls, row):
        return super().__new__(cls, *tuple(row)[0:5])





from ..lib.date import local_date_to_time
from ..command.args import DATE, NAME, VALUE, DELETE, FORCE, mm, COMMAND, CONSTANTS, SET
from ..squeal.database import add
from ..squeal.tables.constant import Constant
from ..squeal.tables.statistic import StatisticJournal, StatisticName, STATISTIC_JOURNAL_CLASSES


def constants(args, log, db):
    '''
## constants

    > ch2 constants [NAME [DATE]]

Lists constants to stdout.

    > ch2 constants --set NAME [DATE] VALUE

Defines a new entry.  If date is omitted a single value is used for all time
(so any previously defined values are deleted)

    > ch2 constants --delete NAME [DATE]

Deletes an entry.

Names can be matched by SQL patterns.  So FTHR.% matches both FTHR.Run and FTHR.Bike, for example.
In such a case "entry" in the descriptions above may refer to multiple entries.
    '''
    name, date, value, set, delete, force = args[NAME], args[DATE], args[VALUE], args[SET], args[DELETE], args[FORCE]
    with db.session_context() as s:
        if name:
            constants = constants_like(log, s, name)
            if not constants:
                raise Exception('Name "%s" matched no entries (see `%s %s`)' % (name, COMMAND, CONSTANTS))
        else:
            constants = []
        if set:
            # date is the optional entry for set
            if date and not value:
                date, value = None, date
            if not name or not value:
                raise Exception('%s requires name and value' % mm(SET))
            set_constants(log, s, constants, date, value, force)
        elif delete:
            if not name:
                raise Exception('%s requires at least a name' % mm(DELETE))
            if value:
                raise Exception('Do not provide a value when deleting Constants')
            if not date and not force:
                raise Exception('Use %s to delete all entries for a Constant' % mm(FORCE))
            delete_constants(log, s, constants, date)
        else:
            if value:
                raise Exception('Do not provide a value when printing Constants')
            print_constants(log, s, constants, name, date)


def constants_like(log, s, name):
    constants = s.query(Constant).filter(Constant.name.like(name)).order_by(Constant.name).all()
    if not constants:
        all_constants = s.query(Constant).all()
        if all_constants:
            log.info('Available constants:')
            for constants in all_constants:
                log.info('%s - %s' % (constants.statistic_name.name, constants.statistic_name.description))
        else:
            log.error('No constants defined - configure system correctly')
        raise Exception('Constant "%s" is not defined' % name)
    return constants


def set_constants(log, s, constants, date, value, force):
    if not date:
        log.info('Checking any previous values')
        journals = []
        for constant in constants:
            journals += s.query(StatisticJournal).join(StatisticName, Constant). \
                filter(Constant.id == constant.id).all()
        if journals:
            log.info('Need to delete %d ConstantJournal entries' % len(journals))
            if not force:
                raise Exception('Use %s to confirm deletion of prior values' % mm(FORCE))
            for journal in journals:
                s.delete(journal)
    for constant in constants:
        constant.add_value(s, value, date=date)
        log.info('Added value %s at %s for %s' % (value, date, constant.name))
    log.warn('You may want to (re-)calculate statistics')


def delete_constants(log, s, constants, date):
    if date:
        for constant in constants:
            for repeat in range(2):
                journal = s.query(StatisticJournal).join(StatisticName, Constant). \
                    filter(Constant.id == constant.id,
                           StatisticJournal.time == date).one_or_none()
                if repeat:
                    log.info('Deleting %s on %s' % (constant.name, journal.time))
                    s.delete(journal)
                else:
                    if not journal:
                        raise Exception('No entry for %s on %s' % (constant.name, date))
    else:
        for constant in constants:
            for journal in s.query(StatisticJournal).join(StatisticName, Constant). \
                    filter(Constant.id == constant.id).order_by(StatisticJournal.time).all():
                log.info('Deleting %s on %s' % (constant.name, journal.time))
                s.delete(journal)


def print_constants(log, s, constants, name, date):
    if not constants:
        constants = s.query(Constant).order_by(Constant.name).all()
        if not constants:
            raise Exception('No Constants defined - configure system')
    print()
    for constant in constants:
        if not date:
            print('%s: %s' % (constant.name,
                              constant.statistic_name.description
                              if constant.statistic_name.description else '[no description]'))
            if name:  # only print values if we're not listing all
                found = False
                for journal in s.query(StatisticJournal).join(StatisticName, Constant). \
                        filter(Constant.id == constant.id).order_by(StatisticJournal.time).all():
                    print('%s: %s' % (journal.time, journal.value))
                    found = True
                if not found:
                    log.warn('No values found for %s' % constant.name)
            print()
        else:
            journal = constant.at(s, date=date)
            if journal:
                print('%s %s %s' % (constant.name, journal.source.time, journal.value))
            else:
                log.warn('No values found for %s' % constant.name)





import datetime as dt

from bokeh.io import show, output_notebook
from bokeh.layouts import column
from bokeh.models.widgets import Dropdown
from bokeh.plotting import figure


def col_to_boxstats(frame, name):
    '''
    Allow inter-op with matplotlib using pre-calculated stats
    See test_data.py
    '''
    stats = []
    for index, row in frame.iterrows():
        if row[name]:
            stats.append({'label': index,
                          'whislo': row[name][0],
                          'q1': row[name][1],
                          'med': row[name][2],
                          'q3': row[name][3],
                          'whishi': row[name][4]})
    return stats


def bokeh_boxplot(f, col):
    '''
    Generate a boxplot for a column (pandas series) containing a tuple of 5 values
    (index date) as provided by summary statistics,
    '''
    def pick(n):
        def pick(x):
            return x[n] if x else None
        return pick
    q = [col.map(pick(i)) for i in range(5)]
    f.segment(q[0].index, q[0], q[1].index, q[1])
    f.vbar(q[1].index, dt.timedelta(days=20), q[1], q[2], fill_alpha=0)
    f.vbar(q[2].index, dt.timedelta(days=20), q[2], q[3], fill_alpha=0)
    f.segment(q[3].index, q[3], q[4].index, q[4])


def statistic_gui(data):

    statistics = data.statistic_names()
    output_notebook()

    def modify_doc(doc):

        f = figure(plot_width=800, plot_height=400)
        f.xaxis.axis_label = 'Date'
        m = Dropdown(label='Statistic',
                     menu=list(zip(statistics['name'].tolist(), (str(i) for i in statistics.index.tolist()))))
        r = column(m, f)

        def mkplot(index):
            f = figure(plot_width=500, plot_height=500)
            f.xaxis.axis_label = 'Date'
            r.children[1] = f
            row = statistics.iloc[int(index)]
            name = row['name']
            d = data.statistic_journals(name, constraint=row['constraint'], owner=row['owner'])
            f.circle(d.index, d[name], line_width=2, legend=name)
            f.legend.location = 'top_left'

        m.on_change('value', lambda attr, old, new: mkplot(new))
        doc.add_root(r)

    show(modify_doc)




from urwid import Text, WEIGHT, Columns

from . import Field, PAGE_WIDTH
from ...lib.utils import label, em
from ...squeal.tables.statistic import StatisticJournal
from ...stoats.calculate.summary import SummaryStatistics


class SummaryField(Field):

    def __init__(self, log, journal, summary=None, width=1):
        if summary is None:
            raise Exception('Summary without summary')
        self._summary = summary
        super().__init__(log, journal, width=width)

    def _format_value(self, value):
        return self._journal.formatted()

    def _widget(self):
        from urwid import Text
        return Text([label('%s: ' % self._summary), em(self._journal.formatted()), ' '])


def summary_columns(log, s, f, date, schedule, names, format_name=lambda n: n):

    def fill(columns, width):
        while width < PAGE_WIDTH:
            columns.append(Text(''))
            width += 1

    def field_columns(name):
        journals = StatisticJournal.at_interval(s, date, schedule, SummaryStatistics, name, SummaryStatistics)
        for named, journal in enumerate(journals):
            summary, period, name = SummaryStatistics.parse_name(journal.statistic_name.name)
            if not named:
                yield Text([format_name(name)]), 1
            display = SummaryField(log, journal, summary=summary)
            yield (WEIGHT, display.width, f(display.widget())), display.width,

    # layout algo here tweaked for subjective appearance
    columns, width = [], 0
    for name in names:
        if width % 2 and width < PAGE_WIDTH:
            columns.append(Text(''))
            width += 1
        lookahead = sum(x[1] for x in field_columns(name))
        if width and width + lookahead > PAGE_WIDTH:
            fill(columns, width)
            yield Columns(columns)
            columns, width = [], 0
        for next_column, next_width in field_columns(name):
            if width + next_width > PAGE_WIDTH:
                fill(columns, width)
                yield Columns(columns)
                columns, width = [], 0
            columns.append(next_column)
            width += next_width

    if columns:
        fill(columns, width)
        yield Columns(columns)
